/*
	< DanaSah is a chess engine winboard. >
    Copyright (C) <2011>  <Pedro Castro Elgarresta>
	E-mail <pecastro@msn.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Funciones de bsqueda del mejor movimiento
*/

/*Includes*/
#include <stdio.h>		/*prinft, scanf, sprinft, sscanf, fgets...*/
#include <string.h>
#include <stdlib.h>
#include "definiciones.h"
#include "variables.h"
#include "funciones.h"

/*Esta funcin nos va a devolver el mejor movimiento posible para una profundidad
Este movimiento podra ser un movimiento de libro
La funcin llama de forma iterativa al algoritmo alphabeta o en nuestro caso PVS*/
movimiento	MotorPiensa()
{
	movimiento	m = no_move;
	int	i, alpha, beta, contar_legales = 0;
	int contador1, contador2;
	char s[4];
	int de, a, tipo;
	int contar;
    movimiento	lista[256];
	char EPD[100];
	char FROM[10],TO[10];
	char FROM1[20],FROM2[20],TO1[20],TO2[20];
	int moves;
	int mejor1 = -MATE, mejor2 = -MATE;
	int tiempo_usado;
	int movimiento_facil1 = 0, movimiento_facil2 = 0;

	if (estado == BUSCANDO) {
		printf ("0 0 0 0 searching...\n");
		fflush(stdout);
		/*ahora vamos a comprobar si hay un movimiento de libro, si el libro est cargado, se han juegado menos de 30
		movimientos y no se han jugado ms de 4 movimientos fuera del libro*/
		if ((strcmp (variante,"normal") == 0) && libro_cargado && fuera_del_libro < 3 && njugadas < 60) {
			s[0] = '\0';
			Tablero_FEN(EPD);
			FIND_OPENING(EPD,FROM,TO,FROM1,FROM2,TO1,TO2,&moves);
			if (moves == 0)
				fuera_del_libro++;
			else
				fuera_del_libro = 0;
			strcat(s, FROM);
			strcat(s, TO);

			/*calculamos las coordenadas del movimiento, de, a y tipo y comparamos con una lista que generamos de todos
			los posibles movimientos, as sabemos que el movimiento que viene del libro es legal*/
			if (strcmp(s,"") != 0) {
				de = s[0] - 'a';
				de += 8 * (8 - (s[1] - '0'));
				a = s[2] - 'a';
				a += 8 * (8 - (s[3] - '0'));
				tipo = NORMAL;

				if (Comprobar_Tipo_Movimiento(de, a, s) != NINGUNO)
					tipo = Comprobar_Tipo_Movimiento(de, a, s);

				/*generamos una lista con todos los posibles movimientos que se podran jugar*/
				contar = Generar(lista);
				/*recorre todos los movimientos para ver si es legal*/
				for (i = 0; i < contar; i++) {
					if (lista[i].de == de && lista[i].a == a && lista[i].tipo == tipo) {
						if (Hacer_Movimiento(lista[i])) {
							Deshacer();			            /*deshacemos el movimiento aunque sea bueno porque lo enviamos luego*/
							printf ("0 0 0 0 Book move\n");	/*indicamos que es un movimiento del libro*/
							fflush(stdout);
							return lista[i];	            /*movimiento que retornamos y se enviar al GUI*/
						}
						else {
							Deshacer();	                    /*si el movimiento del libro no es legal lo deshacemos*/
							continue;
						}
					}
				}
			}
		}
	}
	else if (estado == PONDERANDO) {
		printf ("0 0 0 0 pondering...\n");
		fflush(stdout);
	}

	aborta = FALSO;
	para_busqueda = FALSO;                              /*inicialmente no paramos la bsqueda*/

	tiempo_inicio = get_ms();					        /*tiempo actual*/

	no_nueva_iteracion = max_tiempo / 2;
	no_nuevo_movimiento = (2 * max_tiempo * quedan) / (quedan + 1/2);
	limite_tiempo = (10 * max_tiempo * quedan) / (quedan + 9);

	ply = 0;					                        /*comenzamos en el ply 0 (otros empiezan en el 1)*/
	nodos = 0;					                        /*iniciamos el nmero de nodos*/
	contador1 = contador2 = 0;	                        /*iniciamos contadores para saber si se ha producido un fallo en la bsqueda*/

	material = piezasblanco+piezasnegro+peonesblanco+peonesnegro;

#if UNDER_CE
	if (material<FINAL)
		cknodos =  0x33;
	else
		cknodos =  0xCCC;
#else
	if (material<FINAL)
		cknodos =  0x400;
	else
		cknodos =  0x10000;
#endif

	/*borramos la pv, historia y los movimientos killer y matekiller*/
	memset(pv, 0, sizeof(pv));
	memset(Historia, 0, sizeof(Historia));
	memset(matekiller, 0, sizeof(matekiller));
	memset(killer1, 0, sizeof(killer1));
	memset(killer2, 0, sizeof(killer2));
	maxhistoria = 0;
	/*se guarda el material que hay en ese momento para utilizarlo luego en la evaluacin*/
	old_piezasblanco = piezasblanco;
	old_piezasnegro = piezasnegro;
	old_peonesblanco = peonesblanco;
	old_peonesnegro = peonesnegro;

	/*definimos 2 ventanas para hacer bsqueda, en lugar de buscar desde -MATE hasta MATE que seran todas
	las posibles evaluaciones, buscamos desde -WINDOW1 a +WINDOW1, asi la bsqueda es ms rpida, si la
	bsqueda sale de estos valores abriremos a una ventana ms amplia y luego hasta -MATE y MATE*/
	#define WINDOW1	100
	#define WINDOW2	200

	/*para la primera profundidad hacemos una bsqueda completa*/
	alpha = -MATE;
	beta = MATE;

	contar_legales = Generar_legales();
	//printf("El nmero de movimientos legales es: %d\n", contar_legales);

	/*llamamos a la bsqueda de una forma interactiva, primero buscamos completo la profundidad 1, luego la 2, 
	etc, esto tiene 2 ventajas, si el tiempo se termina al menos tenemos una bsqueda completa en una profundidad 
	y adems cada nueva profundidad podemos resolverla ms rpido ya que produce una mejor ordenacin de la variante principal*/
	for (i = 1; i <= max_profundidad;) {

		li = i;	                /*guardamos el ply para utilizarlo en el acceso a las bitbases y otros*/
		sigue_Pv = VERDADERO;	/*inicialmente hacemos que siga la variante principa*/

		/*llamamos a la bsqueda, la profundidad que vamos a utilizar ser fracional, PLY = 8,
		las extensiones fraccionales deberan permitir un mejor control*/
		Busca_raiz(alpha, beta, i * PLY, &mejor1, &mejor2);
		valor1[i] = mejor1;
		valor2[i] = mejor2;
		m = pv[0][0];		    /*este es el mejor movimiento para la profundidad i*/

		/*mostrar valoracin*/
#if UNDER_CE
		if (estado == BUSCANDO && i > 5) print_pv(i, valor1[i]);
#else
		print_pv(i, valor1[i]);
#endif

		/*si la bsqueda ha terminado, retornamos el mejor movimiento*/
		if (para_busqueda == VERDADERO) {
			while (ply)
				Deshacer();
			break;
		}

		/*abrimos nuevas ventanas si se produce un fallo en la bsqueda, no estoy seguro si tal como est definida
		las ventanas se podra producir inestablidad, parece funcionar bien y rpido*/
		if (valor1[i] <= alpha) {
			contador1++;
			if (contador1 == 2) {
				alpha = -MATE;
				continue;
			}
			alpha = valor1[i] - WINDOW2;
			continue;
		}

		if (valor1[i] >= beta) {
			contador2++;
			if (contador2 == 2) {
				beta = MATE;
				continue;
			}
			beta = valor1[i] + WINDOW2;
			continue;
		}

		/* si la cosa es clara reducimos el tiempo de bsqueda */
		if (estado == BUSCANDO && i > 5) {
			/* ya tenemos una profundidad mnima para abandonar la bsqueda */
			 aborta = VERDADERO;
			/* un solo movimiento legal */
			if (contar_legales == 1) break;
			/*movimiento fcil*/
			if (valor2[i] =! -MATE) {
				if (valor1[i] - valor2[i] > 150) movimiento_facil2++;
				else if (valor1[i] - valor2[i] > 70) movimiento_facil1++;
				if (movimiento_facil2 > 2) {
					no_nueva_iteracion /= 3;
					movimiento_facil2 = 0;
				}
				else if (movimiento_facil1 > 2) {
					no_nueva_iteracion = no_nueva_iteracion * 2 / 3;
					movimiento_facil2 = 0;
				}
			}
			/* no comenzamos una nueva iteracin si ya hemos consumido al menos la mitad de tiempo */
			tiempo_usado = get_ms() - tiempo_inicio;
			if ((tiempo_usado > no_nueva_iteracion) && (valor1[i] + MARGEN > valor1[i-1])) break;
		}

		alpha = valor1[i] - WINDOW1;       /*configuramos la ventana para el siguiente ciclo*/
	    beta = valor1[i] + WINDOW1;
	    i++;
		contador1 = contador2 = 0;
	}

	return m;       /*devolvemos el mejor movimiento*/
}

/* la siguiente funcin es la encargada de la bsqueda cuando la profundidad es 1, est basado en el algoritmo de la variante principal*/
void Busca_raiz(int alpha, int beta, int depth, int *mejor1, int *mejor2)
{
	int i, jaque, da_jaque, contar = 0, nuMoves = 0, ext = 0, puntos;
	movimiento m, mejor_movimiento = no_move, lista[256];   /*218 es el mximo nmero de combinaciones posibles para un movimiento*/
	
	int bound, htipo;
	movimiento hmov;
	int hacer_nulo;
	int hash_flag = hasfALPHA;

	pv_longuitud[ply] = ply;
	jaque = EstaEnJaque(turno);

	*mejor1 = alpha;
	*mejor2 = alpha;

	/*comprobamos las tablas hash solo para ordenar*/
	bound = probe_hash(&htipo, depth+njugadas, beta, &hmov, &hacer_nulo);
	
	contar = Generar(lista);	    /*genera todos los movimientos para la posicin actual*/

	/*ordenamos primero el movimiento que viene de la hash y luego el de la variante principal*/
	Ordena_hmove_pv(contar, lista, hmov);

	/*recorremos todo el ciclo de movimientos*/
    for (i = 0; i < contar; ++i) {
		OrdenaMovimiento(i, contar, lista);
		if (!Hacer_Movimiento(lista[i]))
			continue;
		nuMoves++;	    /*nmero de movimientos legales*/
		nodos++;

		da_jaque = EstaEnJaque(turno);
		
		/*extensiones*/
		if (jaque) ext += PLY;

		/*para el primer movimiento hacemos una busqueda completa*/
		if (nuMoves == 1) {
			if (depth+ext-PLY < PLY && !da_jaque)
				puntos = - Extiende_capturas(-beta, -alpha);
			else
				puntos = -PVS(-beta, -alpha, depth+ext-PLY, VERDADERO);
		}
		else {
			/*LMR*/
			if (nuMoves >= 4 && depth >= 2*PLY && !ext && !da_jaque && !sigue_Pv && lista[i].valor < (maxhistoria*66/100)) {
				puntos = -PVS(-alpha-1, -alpha, depth-2*PLY, VERDADERO);
			}
			else
				puntos = alpha+1;       /*Hack to ensure that full-depth search is done*/
			if (puntos > alpha) {
				if (depth+ext-PLY < PLY && !da_jaque)
					puntos = - Extiende_capturas(-alpha-1, -alpha);
				else
					puntos = -PVS(-alpha-1, -alpha, depth+ext-PLY, VERDADERO);
				if (puntos > alpha && puntos < beta)
					puntos = -PVS(-beta, -alpha, depth+ext-PLY, VERDADERO);
			}
		}

		Deshacer();

		if (para_busqueda == VERDADERO)
			return;

		if (puntos > *mejor1) {

			*mejor2 = *mejor1;
			*mejor1 = puntos;

			if (puntos > alpha) {
				alpha = *mejor1;
				mejor_movimiento = lista[i];
				/*actualizamos la historia*/
				Historia[mejor_movimiento.de][mejor_movimiento.a] += depth/PLY;
				if (Historia[mejor_movimiento.de][mejor_movimiento.a] > maxhistoria)
					maxhistoria = Historia[mejor_movimiento.de][mejor_movimiento.a];
				if (maxhistoria > 2000000) maxhistoria = 2000000;

				if (puntos >= beta) {
					if (puntos > (MATE - 256))
						matekiller[ply] = lista[i];
					else if ((pieza[lista[i].a] == VACIO) && lista[i].tipo < PROMOCION_A_DAMA) {
						killer2[ply] = killer1[ply];
						killer1[ply] = lista[i];
					}
					/*TTables*/
					store_hash(hasfBETA, depth+njugadas, beta, mejor_movimiento);
					return;	    /*cutoff del alphabeta*/
				}

				hash_flag = hasfEXACT;
				Actualiza_pv(mejor_movimiento);	    /*actualizamos la variante principal*/
			}
		}
	}

	store_hash(hash_flag, depth+njugadas, alpha, mejor_movimiento);
	return;
}

/* la siguiente funcin es la encargada de la bsqueda, es un algoritmo alphabeta, una variante
que se llama variante principal, est basada en un algoritmo de Bruce Moreland */
int	PVS(int alpha, int beta, int depth, int hacer_nulo)
{
#define R	2
	int i, jaque, da_jaque, contar = 0, nuMoves = 0, evalua, ev, total_piezas, amenaza_mate = 0, puntos;
	movimiento m, mejor_movimiento = no_move, lista[256];   /*218 es el mximo nmero de combinaciones posibles para un movimiento*/

    int nodo_pv =  ((beta - alpha) > 1);
	int ext;
	int mejor = -MATE;

	/*TTables*/
	int bound, htipo;
	movimiento hmov;
	int hash_flag = hasfALPHA;

	puntos = alpha;

	if ((nodos % cknodos) == 0) 	        /*comprueba cada un cierto nmero de nodos que tenemos tiempo*/
		checkup(puntos);

	if (para_busqueda == VERDADERO)
		return 0;

	/* mate-distance pruning */
	if (alpha >= MATE - 1) { 
		return alpha; 
	}

	jaque = EstaEnJaque(turno);		/*comprobamos si estamos en jaque*/

	/* cut off, I cannot be mated on time faster than beta */    
	if (beta <= -MATE + 2 && !jaque) { 
		return beta;    
	}

	/*si la posicin se repite 1 vez valora como tablas, aunque no las reclama*/
	if (repeticion () > 0)
		return 0;

	pv_longuitud[ply] = ply;

	/*si estamos en profundidad cero o en nuestro caso menor que un PLY llamamos al buscador de capturas y promociones*/
	/*lo mismo si hemos ido muy profundo*/
	if ((depth < PLY && !jaque) || ply >= MAX_PROFUNDIDAD-1) {
		return Extiende_capturas(alpha, beta);
	}

	/*comprobamos las tablas hash*/
	bound = probe_hash(&htipo, depth+njugadas, beta, &hmov, &hacer_nulo);
	if (bound != valUNKNOWN)  {
		switch (htipo) {
			case hasfEXACT:
				if (!nodo_pv) {
					Actualiza_pv(hmov);
					return bound;
				}
				break;
			case hasfALPHA:
				if (bound <= alpha) {
					Actualiza_pv(hmov);
					return alpha;
				}
				break;
			case hasfBETA:
				if (bound >= beta) {
					Actualiza_pv(hmov);
					return beta;
				}
				break;
			case hasfNULL:
				break;
		}
	}

	if (egbb_is_loaded && alpha < MATE - 256 && beta > -MATE + 256) {
		/*comprobamos el acceso a las bitbases*/
		total_piezas = peonesblancos + caballosblancos + alfilesblancos + torresblancas
						+ damasblancas + peonesnegros + caballosnegros + alfilesnegros + torresnegras + damasnegras + 2;

		if (total_piezas <= men) {
			int probe_depth = (2 * li / 3);
			/* Cortamos el rbol solo despus de una cierta profundidad o si hacemos progresos (capturas, movimientos de pen) */
			if (ply >= probe_depth || jugadas[njugadas-1].captura != VACIO || pieza[jugadas[njugadas-1].m.a] == PEON) {
				/* Probamos bitbases en las ramas solo si estn cargadas en RAM */
				if (ply <= probe_depth || (egbb_load_type >= 1 && total_piezas <= 4) || egbb_load_type == 3) {
					bound = 0;
					if (probe_bitbases(&bound) == 1) {
						/* Preferimos ganadas cerca de la raz */
						if (bound > 0)
							bound +=VAL_BITBASE - ply;
						else if (bound < 0)
							bound += ply - VAL_BITBASE;
						return bound;
					}
				}
			}
		}
	}

	ev = Evaluar(alpha, beta);

	/*razoring*/
	evalua = ev + 125;
	if (!jaque && !nodo_pv && hmov.a == 0 && hmov.de == 0 && evalua < beta) {
		if (depth <= PLY) {
			puntos = Extiende_capturas(alpha, beta);
			return MAX(puntos, evalua);
		}
		evalua += 175;
		if (evalua < beta && depth <= 3*PLY) {
			puntos = Extiende_capturas(alpha, beta);
			//if (puntos < beta)
				return MAX(puntos, evalua);
			goto FIN_RAZORING;
		}
		evalua += 200;
		if (evalua < beta && depth <= 4*PLY) {
			puntos = Extiende_capturas(alpha, beta);
			//if (puntos < beta)
				return MAX(puntos, evalua);
			goto FIN_RAZORING;
		}
		evalua += 400;
		if (evalua < beta && depth <= 5*PLY) {
			puntos = Extiende_capturas(alpha, beta);
			//if (puntos < beta)
				return MAX(puntos, evalua);
		}
	}

FIN_RAZORING:

	/*probamos un movimiento nulo, solo lo podemos hacer si nos deja despus de comprobar las tablas hash, sino estamos
	en jaque (ya que nos comeran el rey) y si el lado que mueve tiene al menos una pieza*/
	if (hacer_nulo && !jaque && !sigue_Pv && beta > -MATE + ply && ev >= beta) {

		if ((turno == BLANCO && piezasblanco >= VALOR_CABALLO) || (turno == NEGRO && piezasnegro >= VALOR_CABALLO)) {

			material = piezasblanco+piezasnegro+peonesblanco+peonesnegro;

			/*nosotros utilizamos un movimiento nulo ms agresivo que el resto de programadores, nos da rapidez, no
			seguro si podemos perder algo*/
			depthnull = depth - (1+R+(material>MEDIOJUEGO)+(depth>6*PLY)+(depth>11*PLY)+(depth>16*PLY))*PLY;

			/*hacemos el movimiento nulo, actualizamos la llave hash*/
			jugadas[njugadas].m = no_move;
			jugadas[njugadas].captura = VACIO;
			jugadas[njugadas].enroque = enroque;
			jugadas[njugadas].alpaso = alpaso;
			jugadas[njugadas].regla50 = regla50;
			jugadas[njugadas].hash = hash;
			if (alpaso != -1)
				hash ^= hash_alpaso[alpaso];
			alpaso = -1;
			regla50++;
			ply++;
			njugadas++;
			hash ^= hash_turno[turno];
			turno = 1 - turno;
			hash ^= hash_turno[turno];

			if (depthnull < PLY) 
				puntos = - Extiende_capturas(-beta, -beta+1);
			else
				puntos = -PVS(-beta, -beta+1, depthnull, FALSO);

			/*deshacemos el movmiento nulo*/
			turno = 1 - turno;
			njugadas--;
			ply--;
			m = jugadas[njugadas].m;
			enroque = jugadas[njugadas].enroque;
			alpaso = jugadas[njugadas].alpaso;
			regla50 = jugadas[njugadas].regla50;
			hash = jugadas[njugadas].hash;

			/*verificamos el movimiento nulo si estamos en el final del juego*/
			if (depth >= 6*PLY) {
				if (puntos >= beta) {
					if ((turno == BLANCO && (piezasblanco <= (VALOR_DAMA + VALOR_CABALLO)))
						|| (turno == NEGRO && (piezasnegro <= (VALOR_DAMA + VALOR_CABALLO)))) {

						depthnull = depth - 5*PLY;

						puntos = PVS_no_completo(alpha, beta, depthnull);
					}
				}
			}

			if (puntos >= beta) {
				store_hash(hasfBETA, depth+njugadas, beta, mejor_movimiento);
				return puntos;	/*cutoff*/
			}
			if (puntos <= (-MATE+ply+2))
				amenaza_mate = 1;
		}
	}

    /*Internal Iterative Deepening. Si no tenemos un movimiento proporcionado por las tablas hash y no seguimos
	la variante principal y la profundidad es suficiente, probamos a hacer una bsqueda reducida que nos de un
	movimiento candidato*/
	if (hmov.a == 0 && hmov.de == 0 && nodo_pv && depth >= 8*PLY) {
		puntos = PVS(alpha, beta, depth/2, VERDADERO);
		hmov = pv[ply][ply];
	}

	contar = Generar(lista);	    /*genera todos los movimientos para la posicin actual*/

	/*ordenamos primero el movimiento que viene de la hash y luego el de la variante principal*/
	Ordena_hmove_pv(contar, lista, hmov);

	/*recorremos todo el ciclo de movimientos*/
    for (i = 0; i < contar; ++i) {
		OrdenaMovimiento(i, contar, lista);
		if (!Hacer_Movimiento(lista[i]))
			continue;
		nuMoves++;	    /*nmero de movimientos legales*/
		nodos++;

		da_jaque = EstaEnJaque(turno);

		/*extensiones*/
		ext = 0;
		material = piezasblanco+piezasnegro+peonesblanco+peonesnegro;
		if (nodo_pv) {
			if (jaque) ext += PLY;
			else if (amenaza_mate && ply <= 2*li) ext += PLY;
			else if (jugadas[njugadas-1].captura != VACIO && jugadas[njugadas-2].captura != VACIO
				&& jugadas[njugadas-1].m.a == jugadas[njugadas-2].m.a
				&& abs(valor_pieza[jugadas[njugadas-1].captura] - valor_pieza[jugadas[njugadas-2].captura] <= VALOR_PEON/2))
				ext += PLY;
			/*peon en septima*/
			else if (material < FINAL && pieza[jugadas[njugadas-1].m.a] == PEON
				&& (FILA(jugadas[njugadas-1].m.a) == 1 || FILA(jugadas[njugadas-1].m.a) == 6))
				ext += PLY;
			/*peon en sexta*/
			else if (material < FINAL && pieza[jugadas[njugadas-1].m.a] == PEON
				&& (FILA(jugadas[njugadas-1].m.a) == 2 || FILA(jugadas[njugadas-1].m.a) == 5))
				ext += PLY/2;
			if (material < FINAL && jugadas[njugadas-1].captura > PEON && jugadas[njugadas-1].captura < REY) {
				if (piezasblanco == 0 && piezasnegro == 0) {        /*entrando en un final de peones*/
					switch(jugadas[njugadas-1].captura) {
						case DAMA: ext = 2*PLY;
							break;
						case TORRE: ext = 3*PLY/2;
							break;
						default: ext = PLY;
							break;
					}
				}
			}
		}
		else {
			if (jaque) ext += PLY/2;
		}

		/*para el primer movimiento hacemos una busqueda completa*/
		if (nuMoves == 1) {
			if (depth+ext-PLY < PLY && !da_jaque)
				puntos = - Extiende_capturas(-beta, -alpha);
			else
				puntos = -PVS(-beta, -alpha, depth+ext-PLY, VERDADERO);
		}
		else {
			/*probamos a hacer LMR si no hay extensiones, si hay un nmero mnimo de movimientos legales, si tenemos
			una determinada profundidad, si el valor del movmiento no es grande, que no sea el movimiento de la tabla
			hash, principal variante, matekiller, capturas, promociones y killers y que su historia no sea elevada, si
			todo esto se cumple probamos a reducir la profundidad*/
			if (nuMoves >= 4 && depth >= 2*PLY && !ext && !da_jaque && !sigue_Pv && lista[i].valor < (maxhistoria*66/100)) {
				if (nuMoves > 8 && depth >= 3*PLY && lista[i].valor < (maxhistoria*33/100) && material > MEDIOJUEGO)
					puntos = -PVS(-alpha-1, -alpha, depth-3*PLY, VERDADERO);	//reducimos en 2
				else
					puntos = -PVS(-alpha-1, -alpha, depth-2*PLY, VERDADERO);	//reducimos en 1
			}
			else
				puntos = alpha+1;       /*Hack to ensure that full-depth search is done*/
			if (puntos > alpha) {
				if (depth+ext-PLY < PLY && !da_jaque)
					puntos = - Extiende_capturas(-alpha-1, -alpha);
				else
					puntos = -PVS(-alpha-1, -alpha, depth+ext-PLY, VERDADERO);
				if (puntos > alpha && puntos < beta)
					puntos = -PVS(-beta, -alpha, depth+ext-PLY, VERDADERO);
			}
		}

		Deshacer();

		if (para_busqueda == VERDADERO)
			return mejor;

		if (puntos > mejor) {
			mejor = puntos;
			mejor_movimiento = lista[i];

			if (mejor > alpha) {
				/*actualizamos la historia*/
				Historia[mejor_movimiento.de][mejor_movimiento.a] += depth/PLY;
				if (Historia[mejor_movimiento.de][mejor_movimiento.a] > maxhistoria)
					maxhistoria = Historia[mejor_movimiento.de][mejor_movimiento.a];
				if (maxhistoria > 2000000) maxhistoria = 2000000;

				if (mejor >= beta) {
					if (mejor > (MATE - 256))
						matekiller[ply] = lista[i];
					else if ((pieza[lista[i].a] == VACIO) && lista[i].tipo < PROMOCION_A_DAMA) {
						killer2[ply] = killer1[ply];
						killer1[ply] = lista[i];
					}
					/*TTables*/
					store_hash(hasfBETA, depth+njugadas, beta, mejor_movimiento);
					return mejor;	    /*cutoff del alphabeta*/
				}
				alpha = mejor;
				hash_flag = hasfEXACT;
				Actualiza_pv(mejor_movimiento);	    /*actualizamos la variante principal*/
			}
		}
	}

	if (nuMoves == 0) {
		if (jaque) mejor = -MATE + ply; //mate
		else mejor = 0; // ahogado
	}
	else if (regla50 >= 100)
		return 0;

	/*TTables*/
	store_hash(hash_flag, depth+njugadas, mejor, mejor_movimiento);
	return mejor;
}

/*utilizada cuando hacemos una verificacin del movimiento nulo en el final*/
int	PVS_no_completo(int alpha, int beta, int depth)
{
	int             i, contar, puntos;
	movimiento		lista[256];     /*218 es el mximo nmero de combinaciones posibles para un movimiento*/
	movimiento		hmov = no_move;
	int mejor = -MATE;

	puntos = alpha;

	if ((nodos % cknodos) == 0)
		checkup(puntos);

	if (para_busqueda == VERDADERO)
		return 0;

	pv_longuitud[ply] = ply;

	if (ply >= MAX_PROFUNDIDAD-1)
		return Evaluar(alpha, beta);

	contar = Generar(lista);	/*genera todos los movimientos para la posicin actual*/

	Ordena_hmove_pv(contar, lista, hmov);

	/*recorremos todo el ciclo de movimientos*/
    for (i = 0; i < contar; ++i) {
		OrdenaMovimiento(i, contar, lista);
		if (!Hacer_Movimiento(lista[i]))
			continue;

		nodos++;

		puntos = -PVS(-beta, -alpha, depth-PLY, VERDADERO);     /*nuestra puntuacin es la del contrario con signo contrario*/

		Deshacer();

		if (puntos > mejor) {
			mejor = puntos;
			if (mejor > alpha) {
				if (mejor >= beta)
					return beta;
				alpha = mejor;
				Actualiza_pv(lista[i]);
			}
		}
	}
	return mejor;
}

/*extensiones es similar a alphabeta salvo en que solo busca capturas y promociones, no lleva profundidad, esta funcin
siver para evitar el efecto horizonte*/
int Extiende_capturas(int alpha, int beta)
{
	int i, contar2, delta, puntos;
	movimiento lista2[256], hmov = no_move;

	puntos = alpha;

	if ((nodos % cknodos) == 0)
		checkup(puntos);

	if (para_busqueda == VERDADERO)
		return 0;

	pv_longuitud[ply] = ply;

	if (ply >= MAX_PROFUNDIDAD-1)
		return Evaluar(alpha, beta);

	puntos = Evaluar(alpha, beta);

	if (puntos >= beta)
        return beta;

	if (puntos > alpha)
        alpha = puntos;

    /*delta prunning*/
	material = piezasblanco+piezasnegro+peonesblanco+peonesnegro;
	if (material > FINAL) {
		Evalua_material();
        delta = alpha - 200 - (turno == BLANCO ? (puntosblanco-puntosnegro) : (puntosnegro-puntosblanco));
    } else {
        delta = 0;
    }

    if (delta > VALOR_DAMA) {
        return puntos;
    } else if (delta > VALOR_TORRE) {
        if (turno == BLANCO) {
            if (!damasnegras)
                return puntos;
        } else {
            if (!damasblancas)
                return puntos;
        }

    } else if (delta > VALOR_ALFIL) {
        if (turno == BLANCO) {
            if (!damasnegras && !torresnegras)
                return puntos;
        } else {
            if (!damasblancas && !torresblancas)
                return puntos;
        }
    } else if (delta > VALOR_PEON) {
        if (turno == BLANCO) {
            if (!damasnegras && !torresnegras && !caballosnegros && !alfilesnegros)
                return puntos;
        } else {
            if (!damasblancas && !torresblancas && !caballosblancos && !alfilesblancos)
                return puntos;
        }
    }

	contar2 = Generar_Capturas(lista2);

	Ordena_hmove_pv(contar2, lista2, hmov);

    for (i = 0; i < contar2; ++i) {
		OrdenaMovimiento(i, contar2, lista2);
		if (lista2[i].tipo < PROMOCION_A_DAMA) {
			/* mala captura y delta prunning */
			if (lista2[i].valor == -1 || abs(valor_pieza[pieza[lista2[i].a]]) <= delta)
				continue;
		}
		if (!Hacer_Movimiento(lista2[i]))
			continue;

		nodos++;

		puntos = -Extiende_capturas(-beta, -alpha);
		Deshacer();

		if (puntos >= beta)
			return beta;

		if (puntos > alpha){
			alpha = puntos;
			Actualiza_pv(lista2[i]);
		}
	}
	return alpha;
}

/*ordena el movimiento hash en primer lugar, luego el movimiento que sigue la variante principal anterior*/
void Ordena_hmove_pv(int cuenta, movimiento *lista, movimiento hmove)
{
	int i;

	if (!sigue_Pv) {
		for (i = 0; i < cuenta; ++i) {
			if (hmove.de == lista[i].de && hmove.a == lista[i].a) {
				lista[i].valor += 20000000;
				return;
			}
		}
	}
	else {
		for(i = 0; i < cuenta; ++i) {
			if (hmove.de == lista[i].de && hmove.a == lista[i].a) {
				lista[i].valor += 20000000;
				break;
			}
		}
		sigue_Pv = FALSO;
		for(i = 0; i < cuenta; ++i) {
			if (lista[i].de == pv[0][ply].de && lista[i].a == pv[0][ply].a) {
				sigue_Pv = VERDADERO;
				lista[i].valor += 10000000;
				break;
			}
		}
	}
}

/*ponemos en primer lugar el movimiento con mayor valor*/
void OrdenaMovimiento(int de, int cuenta, movimiento *lista)
{
	int i, j, Max;
	movimiento Tmp;

	/*cerca del maximo*/
	Max = lista[de].valor;
	j = de;
	for (i = (de+1); i < cuenta; i++) {
		if (lista[i].valor > Max) {
			Max = lista[i].valor;
			j = i;
		}
	}
	if (Max != lista[de].valor) {
		Tmp = lista[de];
		lista[de] = lista[j];
		lista[j] = Tmp;
	}
}

/*actualizamos la variante principal con un movimiento*/
void Actualiza_pv(movimiento pvmov)
{
	int j;

	pv[ply][ply] = pvmov;
	for (j = ply + 1; j < pv_longuitud[ply + 1]; ++j)
		pv[ply][j] = pv[ply + 1][j];
	pv_longuitud[ply] = pv_longuitud[ply + 1];
}

/*comprueba si una posicin se repite*/
int repeticion()
{
	int i;
	int r = 0;

	for (i = njugadas-regla50; i < njugadas; ++i)
		if (jugadas[i].hash == hash)
			++r;
	return r;
}

void checkup(int puntos)
{
	char linea[256], comando[256];	    /*para leer la informacin del GUI y saber que comando llega*/
	int tiempo_usado;

	tiempo_usado = get_ms() - tiempo_inicio;

	if (estado == BUSCANDO && aborta) {
		if (tiempo_usado >= limite_tiempo) {
			para_busqueda = VERDADERO;
		}
		else if  ((tiempo_usado >= no_nuevo_movimiento) && (puntos + MARGEN > valor1[li-1])) {
			para_busqueda = VERDADERO;
		}
	}
	else if (estado == PONDERANDO) {
		if (bioskey()) {
			para_busqueda = VERDADERO;
		}
	}
	else if (estado == ANALIZANDO) {
		if (bioskey()) {
#if UNDER_CE
			gets(linea);
#else
			fgets(linea, 256, stdin);
#endif
			sscanf(linea, "%s", comando);
			if (strcmp(comando, "exit") == 0) {
				para_busqueda = VERDADERO;
			}
		}
	}
}
