/*
	< DanaSah is a chess engine winboard. >
    Copyright (C) <2011>  <Pedro Castro Elgarresta>
	E-mail <pecastro@msn.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Funcin xboard(), sustituye a la funcin principal si recibimos el parmetro xboard,
gracias a estos comandos, podemos hacer jugar a DanaSah con interfaces de usuario como
Winboard, Arena, Fritz, Chessmaster, etc.
Para comprender dicho protocolo debes examinar la siguiente pgina:
http://www.tim-mann.org/xboard/engine-intf.html
*/

/*Includes*/
#if UNDER_CE
	#include <windows.h>	/*necesaria para MessageBox*/
#endif
#include <stdio.h>		/*prinft, scanf, sprinft, sscanf, fgets...*/
#include <string.h>
#include <stdlib.h>		/*malloc, free, exit ...*/
#include "definiciones.h"
#include "variables.h"
#include "funciones.h"

void xboard() /*protocolo xboard*/
{
	char linea[256], comando[256], comando2[256];	/*para leer la informacin del GUI y saber que comando llega*/
	char args[4][64];
	int post;		                                /*utilizada para imprimir la linea principal o no (no utilizada de momento)*/
	int entrada = 0;                                /*recibe el nmero de movimientos para el control de tiempo*/
	char minutos[4];	                            /*recibe el nmero de minutos base para el control de tiempo*/
	int incremento;	                                /*recibe el nmero de segundos de incremento para el control de tiempo*/
	int contar;		                                /*numero de movimientos posibles en una posicin*/
	movimiento lista[256], mejormovimiento;	        /*lista de movimientos posibles y mejormovimiento*/
	int error;

#if !UNDER_CE
	setbuf(stdout, NULL);
#endif

	printf("feature done=0\n");

	/*cargamos el libro de aperturas*/
	if (usa_libro) {

		error=INITIALIZE(librotorneo,libroprincipal);   /*Initialize (one time call)*/

		if (error == 0) {
			libro_cargado = VERDADERO;
#if UNDER_CE
			puts ("Rebel book found!");
			//MessageBox (NULL, STRING ("Rebel book found!\r\n"), _T("Engine DanaSah 4.24"), MB_OK);
#else
			printf("Rebel book loaded!\n");
#endif			
		}
		else if (error == 1) {
			libro_cargado = VERDADERO;
#if UNDER_CE
			puts ("Rebel book found but tourbook is missed, using only mainbook");
			MessageBox (NULL, STRING ("Rebel book found but tourbook is missed, using only mainbook\r\n"), _T("Engine DanaSah 4.24"), MB_OK);
#else
			printf("Rebel book loaded but tourbook is missed, using only mainbook\n");
#endif			
		}
		else if (error == 2) {
#if UNDER_CE
			puts ("Missing obliged file(s)random1.bin and/or random2.bin");
			MessageBox (NULL, STRING ("Missing obliged file(s) random1.bin and/or random2.bin\r\n"), _T("Engine DanaSah 4.24"), MB_OK);
#else
			printf("Missing obliged file(s)random1.bin and/or random2.bin. Rebel book not loaded!\n");
#endif			
		}
		else if (error == 3) {
#if UNDER_CE
			puts ("Missing obliged file mainbook");
			MessageBox (NULL, STRING ("Missing obliged file mainbook\r\n"), _T("Engine DanaSah 4.24"), MB_OK);
#else
			printf("Missing obliged file mainbook. Rebel book not loaded!\n");
#endif			
		}
		else {
#if UNDER_CE
			puts ("Error loading Rebel book!");
			MessageBox (NULL, STRING ("Error loading Rebel book!\r\n"), _T("Engine DanaSah 4.24"), MB_OK);
#else
			printf("Error loading Rebel book!\n");
#endif			
		}
	}

	/*cargamos las bitbases*/
	egbb_cache_size = (egbb_cache_size * 1024 *1024);
	egbb_is_loaded = LoadEgbbLibrary(egbb_path, egbb_cache_size);

	/*bucle esperando recibir un comando del GUI*/
	for (;;) {
		fflush(stdout);	                            /*enviamos el contenido del buffer a la pipe de xboard evitando el buffering*/

        if (turno == motor) {                       /*turno del programa*/
			estado = BUSCANDO;
			mejormovimiento = MotorPiensa();	    /*piensa la jugada*/
			Hacer_Movimiento(mejormovimiento);	    /*guarda el movimiento en la lista de jugadas*/

			Enviar_Movimiento(mejormovimiento);     /*envia el movimiento al GUI*/
			contar = Generar(lista);		  	    /*genera los movimientos del contrario para saber si ha recibido mate*/
			imprime_resultado(contar, lista);	    /*si es final de partida imprime el resultado*/

#if UNDER_CE
			if (max_tiempo > 1000)
				hacer_ponder = VERDADERO;
			else
				hacer_ponder = FALSO;
#else
			hacer_ponder = VERDADERO;
#endif
			continue;        
        }
		else {
    		if (ponder && hacer_ponder) {		    /*pondera si el GUI le enva una seal*/
#if UNDER_CE
				if (fuera_del_libro) {				/*lo hacemos si no estamos en el libro*/
					estado = PONDERANDO;
					mejormovimiento = MotorPiensa();    /*piensa la jugada*/
       			}
#else
				estado = PONDERANDO;
				mejormovimiento = MotorPiensa();    /*piensa la jugada*/
#endif
				hacer_ponder = FALSO;			    /*deja de ponderar*/
    			continue;
            }
        }
		/*almacenamos lo que nos llega y lo guardamos en la variable comando*/
#if UNDER_CE
		if (gets(linea) == NULL)
			return;
#else
		if (!fgets(linea, 256, stdin))
			return;
#endif
		if (linea[0] == '\n')
			continue;
		sscanf(linea, "%s", comando);
		/*comprueba si tenemos protocolo winboard 2 y definimos algunas caractersticas*/
		if (strcmp(comando,"protover") == 0){
			printf("feature myname=\"%s\" ping=1 setboard=1 playother=1 draw=0 sigint=0 sigterm=0 variants=\"normal,nocastle,fischerandom\" colors=0 ics=1\n", nombre);
			printf("feature done=1\n");
			continue;
		}
		/*nos indica si las caractersticas definidas son aceptadas o no*/
		if (strcmp(comando, "accepted") == 0) {
			/*printf("feature accepted\n");*/
			continue;
		}
		if (strcmp(comando, "rejected") == 0) {
			printf("feature rejected\n");
			continue;
		}
		/*si llega xboard simplemente continuamos*/
		if (strcmp(comando, "xboard") == 0) {
			continue;
		}
		/*comprueba si hay que empezar nueva partida*/
		if (strcmp(comando, "new") == 0) {
			variante[0] = '\0';
			strcpy(variante,"normal");
			//printf("variante = %s\n", variante);
			strcpy(fenstring, "");
			strcat(fenstring, STARTPOS);
			SetBoard(fenstring);
			motor = NEGRO;
			continue;
		}
		if (strcmp(comando, "variant") == 0) {
			sscanf(linea, "variant %s", comando2);
			variante[0] = '\0';
			strcpy(variante,comando2);
			//printf("variante = %s\n", variante);
			continue;
		}
		/*comando para cerrar el motor */
		if (strcmp(comando, "quit") == 0)
			return;
		if (strcmp(comando, "random") == 0)
			continue;
		/*cambia el color de las piezas entre los jugadores*/
		if (strcmp(comando, "go") == 0) {
				motor = turno;
			continue;
		}
		/*el motor deja de jugar*/
		if (strcmp(comando, "force") == 0) {
			motor = VACIO;
			continue;
		}
		/*asigna colores, el comando playother es ms moderno que white o black*/
		if (strcmp(comando, "playother") == 0) {
			if (turno == BLANCO)
				motor = NEGRO;
			else
				motor = BLANCO;
			continue;
		}
		if (strcmp(comando, "white") == 0) {
			turno = BLANCO;
			motor = NEGRO;
			continue;
		}
		if (strcmp(comando, "black") == 0) {
			turno = NEGRO;
			motor = BLANCO;
			continue;
		}
		/*parmetros para configurar el nivel de juego del programa	level 40 5 0 --> 40 movimientos 5 minutos incremento 0
		si el nmero de movimientos es 0 la partida es a finish*/
		if (strcmp(comando, "level") == 0) {
			sscanf(linea, "level %d %s %d", &entrada, minutos, &incremento);
			mps = entrada;	    /*nos interesa saber cuantos movimientos para el control*/
			inc = incremento * 1000;
			continue;
		}
		/*configura el nivel de juego indicando cuantos segundos por movimiento	st 10 --> 10 segundos por jugada*/
		if (strcmp(comando, "st") == 0) {
			sscanf(linea, "st %d", &max_tiempo);
			max_tiempo *= 1000;		/*internamente trabajamos con milisegundos*/
			quedan = 1;
			/*con una profundidad demasiada profunda quizs podra tener problemas para salir rpido y perder en el tiempo*/
			if (max_tiempo < 3000)
				max_profundidad = MAX_PROFUNDIDAD/2;
			else
				max_profundidad = MAX_PROFUNDIDAD;
			continue;
		}
		/*configura el nivel de juego indicando la profundidad a jugar sd 8 --> juega hasta profundidad principal 8*/
		if (strcmp(comando, "sd") == 0) {
			sscanf(linea, "sd %d", &max_profundidad);
			max_tiempo = 1 << 25;	/*es como poner tiempo infinito*/
			quedan = 1;
			continue;
		}
		/*recibimos del GUI el tiempo que nos queda*/
		if (strcmp(comando, "time") == 0) {
			sscanf(linea, "time %d", &max_tiempo);
			/*pasamos a milisegundos que es como trabajamos internamente*/
			max_tiempo *= 10;
			/*partidas a finish, suponemos que quedan 40 movimientos para terminar la partida*/
			if (mps == 0 && inc == 0) {
#if UNDER_CE
				if (njugadas < 40) {
					max_tiempo /= 40;
					quedan = 40
				}
				else if (njugadas < 80) {
					max_tiempo /= 35;
					quedan = 35
				}
				else if (njugadas < 120) {
					max_tiempo /= 30;
					quedan = 25;
				}
				else {
					max_tiempo /= 25;
					quedan = 25;
				}
#else
				max_tiempo /= 40;
				quedan = 40;
#endif
			}
			/*partidas con incremento de tiempo*/
			else if (mps == 0 && inc > 0) {
				if (max_tiempo < 3 * inc)
					max_tiempo = inc * time_percentage / 100;
				else
					max_tiempo = max_tiempo / 30 + inc;
				quedan = 30;
			}
			/*si hay un control de tiempo*/
			else {
				/*movimientos que quedarn para llegar a dicho control
				el contador njugadas se va incrementando tanto para blancas como negras*/
				if (motor == BLANCO)
					quedan = (mps-njugadas/2)-1;
				else
					quedan = (mps-njugadas/2);

				/*quitamos un pequeo tiempo para que no pierda en el ltimo movimiento*/
				if (quedan == 1) {
					if (max_tiempo >= 7500)
						max_tiempo -= 2000;		/*damos 2 segundos de margen*/
					else if (max_tiempo < 1500)
						max_tiempo = max_tiempo * (time_percentage-30) / 100;
					else
						max_tiempo = max_tiempo * time_percentage / 100;
				}
				/*una vez llegado al control de tiempo, lo ponemos de nuevo a cero*/
				if (quedan == 0) {
					max_tiempo = max_tiempo/entrada;
					mps = mps + entrada;
					quedan = entrada;
				}
				/*si estamos en medio de un control*/
				else {
					if (quedan <= 5 && max_tiempo < 3000)
						max_tiempo /= 2;
					/*el tiempo para un movimiento ser el tiempo que queda entre los
					movimientos que quedan para el control*/
					max_tiempo = max_tiempo/quedan;
				}
			}
			if (max_tiempo < 3000)
				max_profundidad = MAX_PROFUNDIDAD/2;
			else
				max_profundidad = MAX_PROFUNDIDAD;
			continue;
		}
		/*tiempo del contrario, no lo utilizamos de momento*/
		if (strcmp(comando, "otim") == 0) {
			continue;
		}
		if (strcmp(comando, "usermove") == 0) {
			continue;
		}
		/*comando para mover ahora*/
		if (strcmp(comando, "?") == 0) {
			if (estado == BUSCANDO) {
				para_busqueda = VERDADERO;
			}
			continue;
		}
		/*control para winboard, recibimos un ping y enviamos un pong*/
		if (strcmp(comando,"ping") == 0){
			int n;
			sscanf(linea, "ping %d", &n);
			printf("pong %d\n",n);
			continue;
		}
		if (strcmp(comando, "draw") == 0) {
			continue;
		}
		if (strcmp(comando, "result") == 0) {
			motor = VACIO;
			njugadas=0;
			continue;
		}
		if (strcmp(comando, "setboard") == 0) {
			strcpy(fenstring, "");
			sscanf(linea, "setboard %s %s %s %s", args[0],args[1],args[2],args[3]);
			strcat(fenstring, args[0]);
			strcat(fenstring, args[1]);
			strcat(fenstring, args[2]);
			strcat(fenstring, args[3]);
			SetBoard(fenstring);
			continue;
		}
		/*viejo comando edit de winboard para configurar una posicin*/
		if (strcmp(comando, "edit") == 0) {
			continue;
		}
		if (strcmp(comando, "hint") == 0) {
			char c;
			movimiento mh;
			mh = pv[0][1];
			if (!(mh.de == 0 && mh.a == 0)) {
				/*letra a imprimir si hay una promocin*/
				switch (mh.tipo) {
					case PROMOCION_A_DAMA:
						c = 'q';
						break;
					case PROMOCION_A_TORRE:
						c = 'r';
						break;
					case PROMOCION_A_ALFIL:
						c = 'b';
						break;
					case PROMOCION_A_CABALLO:
						c = 'n';
						break;
					default:
						c = ' ';
				}
				printf("Hint: %c%d%c%d%c",
					'a' + COLUMNA(mh.de),
					8 - FILA(mh.de),
					'a' + COLUMNA(mh.a),
					8 - FILA(mh.a),
					c);
			}
			continue;
		}
		if (strcmp(comando, "bk") == 0) {
			continue;
		}
		/*deshacemos una jugada*/
		if (strcmp(comando, "undo") == 0) {
			if (njugadas == 0)
				continue;
			Deshacer();
			continue;
		}
		if (strcmp(comando, "remove") == 0) {
			if (njugadas < 2)
				continue;
			Deshacer();
			Deshacer();
			continue;
		}
		/*ponder, pensar con el tiempo del contrario*/
		if (strcmp(comando, "hard") == 0) {
			ponder = VERDADERO;
			continue;
		}
		/*no ponder*/
		if (strcmp(comando, "easy") == 0) {
			ponder = FALSO;
			continue;
		}
		/*imprimimos el pensamiento*/
		if (strcmp(comando, "post") == 0) {
			post = VERDADERO;
			continue;
		}
		/*no imprimir el pensamiento*/
		if (strcmp(comando, "nopost") == 0) {
			post = FALSO;
			continue;
		}
		/*modo de anlisis con GUI*/
		if (strcmp(comando, "analyze") == 0) {
			max_tiempo = 1<<25;
			max_profundidad = MAX_PROFUNDIDAD;
			estado = ANALIZANDO;
			mejormovimiento = MotorPiensa();
			continue;
		}
		if (strcmp(comando, ".") == 0) {
			continue;
		}
		if (strcmp(comando, "name") == 0) {
			continue;
		}
		if (strcmp(comando, "rating") == 0) {
			continue;
		}
		if (strcmp(comando, "ics") == 0) {
			sscanf(linea, "ics %s", comando2);
			servidor[0] = '\0';
			strcpy(servidor,comando2);
			continue;
		}
		if (strcmp(comando, "computer") == 0) {
			continue;
		}
		if (strcmp(comando, "pause") == 0) {
			continue;
		}
		if (strcmp(comando, "resume") == 0) {
			continue;
		}
#if UNDER_CE
		if (strcmp(comando, "log") == 0) {
			continue;
		}
		if (strcmp(comando, "output") == 0) {
			continue;
		}
		if (strcmp(comando, "unrated") == 0) {
			continue;
		}
		if (strcmp(comando, "rated") == 0) {
			continue;
		}
#endif

		Comprobar_Movimiento(comando);			/*comprueba si llega un movimiento del usuario y es legal*/
		contar = Generar(lista);		  	    /*genera los movimientos del contrario para saber si ha recibido mate*/
		imprime_resultado(contar, lista);	    /*si es final de partida imprime el resultado*/
		if (salir == VERDADERO){				/*si es final de partida el motor deja de jugar*/
			motor = VACIO;
			njugadas=0;
			continue;
		}
	}
}
